from .enums import RegisterTypes,DataTypes
from .data_value import DataValue
from .register_slice import RegisterSlice
import logging

class Component(object):
    def __init__(self,input_address:int,holding_address:int=-1)->None:
        self.input_address = input_address
        self.input_count = 0
        self.holding_address = holding_address
        self.holding_count = 0
        self.__data_values = None
        self.__input_values = None
        self.__holding_values = None
        
    def _initialize(self):
        """
        Initializes the absolute addresses of the DataValues and the count of the Component
        """
        for _,value in self.__get_data_values():
            if value.register_type == RegisterTypes.Input:
                value._absolut_address = self.input_address
            else:
                value._absolut_address = self.holding_address
          
        #Dynamically calcuate how many registers have to be read  
        if len(self.__get_input_values()) > 0:    
            _,last_input_value = self.__get_input_values()[-1]
            self.input_count = last_input_value.address+last_input_value.count
        
        if len(self.__get_holding_values()) > 0:
            _,last_holding_value = self.__get_holding_values()[-1]
            self.holding_count = last_holding_value.address+last_holding_value.count
            
        #Calculate the address slices we need to read
        #This is necessary because the modbus protocol can block the read/write of some registers
        #=> if one of these registers is between the start and end address of the read/write we need to skip it
        self.__input_slices = Component._calculate_ranges([v for (n,v) in self.__get_input_values()])
        self.__holding_slices = Component._calculate_ranges([v for (n,v) in self.__get_holding_values()])
        return self
    
    @property
    def holding_slices(self)->list[RegisterSlice]:
        """
        Returns the address slices of the holding registers
        """
        return self.__holding_slices
    
    @property
    def input_slices(self)->list[RegisterSlice]:
        """
        Returns the address slices of the input registers
        """
        return self.__input_slices
    
    @staticmethod      
    def _calculate_ranges(datavalues:list[DataValue])->list[RegisterSlice]:
        slices = []
        if len(datavalues) < 1:
            return slices
        current_slice = RegisterSlice(datavalues[0].get_absolute_address(),datavalues[0].address,datavalues[-1].address+datavalues[-1].count)
        for i,datavalue in enumerate(datavalues[:-1]):
            if datavalue.address+datavalue.count != datavalues[i+1].address:
                #A gap was found => define the range from the start of the current slice to the end of the current address+count
                current_slice.count = (datavalue.address+datavalue.count)-current_slice.relative_address
                slices.append(current_slice)
                #the new count is the distance between the next address and the last address
                new_count = (datavalues[-1].address+datavalues[-1].count)-datavalues[i+1].address
                current_slice = RegisterSlice(datavalues[i+1].get_absolute_address(),datavalues[i+1].address,new_count)
        slices.append(current_slice)
        return slices
    @property
    def has_input_address(self)->bool:
        return self.input_address >= 0 and self.input_count > 0
    
    @property
    def has_holding_address(self)->bool:
        return self.holding_address >= 0 and self.holding_count > 0
      
    def __get_data_values(self)->list[tuple[str,DataValue]]:
        """
        Returns all DataValues of this Component
        """
        if self.__data_values is None:
            self.__data_values = [(k,v) for k,v in self.__dict__.items() if isinstance(v,DataValue)]
        return self.__data_values
    
    def __get_input_values(self)->list[tuple[str,DataValue]]:
        """
        Get sorted list of all DataValues with RegisterType Input
        """
        if self.__input_values is None:
            self.__input_values = [(k,v) for k,v in self.__get_data_values() if v.register_type == RegisterTypes.Input]
            self.__input_values = sorted(self.__input_values,key=lambda item: item[1].address)
        return self.__input_values
    
    def __get_holding_values(self)->list[tuple[str,DataValue]]:
        """
        Get sorted list of all DataValues with RegisterType Holding
        """
        if self.__holding_values is None:
            self.__holding_values = [(k,v) for k,v in self.__get_data_values() if v.register_type == RegisterTypes.Holding]
            self.__holding_values = sorted(self.__holding_values,key=lambda item: item[1].address)
        return self.__holding_values
    
    @staticmethod
    def __unsigned_to_signed(n:int, byte_count:int)->int:
        return int.from_bytes(
            n.to_bytes(byte_count, "little", signed=False), "little", signed=True
        )
        
    def parse(self,data:list[int],type:RegisterTypes)->bool:
        """
        Dynamically assignes the values to the DataValues of this Component
        """
        if len(data) != (self.input_count if type == RegisterTypes.Input else self.holding_count):
            logging.error(f"Data length does not match the expected length of {self.input_count if type == RegisterTypes.Input else self.holding_count} for {self.__class__.__name__}")
            return False
        
        encountered_error = False
        for name,value in self.__get_input_values() if type == RegisterTypes.Input else self.__get_holding_values():
            try:
                # Multi-register values (UINT32, INT32)
                if value.count == 2:
                    _value = (data[value.address] << 16) + data[value.address+ 1]
                else:
                    _value = data[value.address]
                 
                # Datatype
                if value.type == DataTypes.INT:
                    _value = Component.__unsigned_to_signed(_value, value.count * 2)
                    
                #Store
                value.value = _value   
            except:
                logging.exception(f"Error while parsing {name} of {self.__class__.__name__}")
                encountered_error = True
        return not encountered_error
    
    def __repr__(self) -> str:
        s = ["="*12]
        s.append(f"{self.__class__.__name__}")
        s.append("="*12)
        if self.has_input_address:
            s.append("---Input:")
            for name,value in self.__get_input_values():
                s.append(f"{name}| raw:{value.value} scaled:{value.scaled_value}")
        if self.has_holding_address:
            s.append("---Holding:")
            for name,value in self.__get_holding_values():
                s.append(f"{name} | raw:{value.value}")     
        return "\n".join(s)


from .components.heating_circuit import *
from .components.boiler import *
from .components.heat_pump import *
from .components.buffer import *
from .components.pellets_boiler import *
from .components.photovoltaic import *
from . import Systems


class ComponentFactory:
    @staticmethod
    def heating_circuit(system:Systems)->HeatingCircuit:
        if system == Systems.Therminator:
            return TherminatorHeatingCircuit()._initialize()
        return HeatingCircuit()._initialize()
    
    @staticmethod
    def boiler(system:Systems)->Boiler:
        return Boiler()._initialize()
    
    @staticmethod
    def heatpump(system:Systems)->HeatPump:
        return HeatPump()._initialize()
    
    @staticmethod
    def photovoltaic(system:Systems)->Photovoltaic:
        return Photovoltaic()._initialize()
    
    @staticmethod
    def pelletsboiler(system:Systems)->PelletsBoiler:
        return PelletsBoiler()._initialize()
    
    @staticmethod
    def buffer(system:Systems)->Buffer:
        if system == Systems.Therminator:
            return TherminatorBuffer()._initialize()
        else:
            return Buffer()._initialize()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysolarfocus', 'pysolarfocus.components', 'pysolarfocus.components.base']

package_data = \
{'': ['*']}

install_requires = \
['pymodbus==2.5.3']

setup_kwargs = {
    'name': 'pysolarfocus',
    'version': '2.0.1',
    'description': 'Unofficial, local Solarfocus client',
    'long_description': "# pysolarfocus: Python Client for Solarfocus eco<sup>_manager-touch_</sup>\n\n## What's Supported \n\n### Software Version\n\nThis integration has been tested with Solarfocus eco<sup>manager-touch</sup> version `21.040`.\n\n### Solarfocus Components\n\n| Components | Supported |\n|---|---|\n| Heating Circuit 1 (_Heizkreis_)| :white_check_mark: |\n| Buffer 1 (_Puffer_) | :white_check_mark: |\n| Solar (_Solar_)| :x:|\n| Boiler 1 (_Boiler_) | :white_check_mark: |\n| Heatpump (_Wärmepumpe_) | :white_check_mark: |\n| Biomassboiler (_Kessel_) | :white_check_mark: | \n\n_Note: The number of supported Heating Circuits, Buffers, and Boilers could be extended in the future_\n\n## Usage\n\n```python\nfrom pymodbus.client import ModbusTcpClient as ModbusClient\nfrom pysolarfocus import SolarfocusAPI,PORT,Systems\n\n# Create a Modbus client\nclient = ModbusClient(IP, port=PORT)\nclient.connect()\n\n# Create the Solarfocus API client\nsolarfocus = SolarfocusAPI(client, Systems.Vampair)\n\n# Fetch the values\nsolarfocus.update()\n\n# Print the values\nprint(solarfocus.buffer)\nprint(solarfocus.heating_circuit)\n```\n",
    'author': 'Jeroen Laverman',
    'author_email': 'jjlaverman@web.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lavermanjj/pysolarfocus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .base.component import Component
from .base.enums import DataTypes
from .base.data_value import DataValue

class PelletsBoiler(Component):
    def __init__(self) -> None:
        super().__init__(2400)
        self.temperature = DataValue(address=0,multiplier=0.1)
        self.status  = DataValue(address=1,type=DataTypes.UINT)
        self.message_number = DataValue(address=4)
        self.door_contact = DataValue(address=5)
        self.cleaning = DataValue(address=6)
        self.ash_container = DataValue(address=7)
        self.outdoor_temperature = DataValue(address=8,multiplier=0.1)
        self.boiler_operating_mode = DataValue(address=9)
        self.octoplus_buffer_temperature_bottom = DataValue(address=10,multiplier=0.1)
        self.octoplus_buffer_temperature_top = DataValue(address=11,multiplier=0.1)
        self.log_wood = DataValue(address=12,type=DataTypes.UINT)


"""Constants for pysolarfocs"""

# Values provided by Solarfocus manual
SLAVE_ID = 1

HEATING_STATE = {
    0: "Heizkreis ist ausgeschaltet",
    1: "Absenkbetrieb",
    2: "Heizbetrieb",
    3: "Ferienbetrieb",
    4: "Estrichprogramm",
    5: "Frostschutzbetrieb",
    6: "Kaminkehrer",
    7: "Heizkreis nicht freigeschaltet",
    8: "Wärmeableitung",
    9: "Außenabschalttemperatur Heizbetrieb erreicht",
    10: "Raumsolltemperatur Heizbetrieb erreicht",
    11: "Trinkwasserspeichervorrang ist aktiv",
    12: "Dauerheizbetrieb",
    13: "Dauerabsenkbetrieb",
    14: "Aussenfühlerunterbrechung",
    15: "min. Energiequellentemperatur unterschritten",
    16: "Vorlauffühler defekt",
    17: "min. Energiequellentemperatur unterschritten, Frostschutzbetrieb",
    18: "Testlauf Pumpe ist aktiv",
    19: "Partybetrieb",
    20: "Begrenzungsthermostat ist offen",
    21: "Pumpen Nachlauf",
    22: "Defrost",
    23: "Kühlbetrieb",
    24: "Kühlen hat Vorrang",
    25: "Heizen hat Vorrang",
    26: "Pool hat Vorrang",
    27: "Außenabschalttemperatur Absenkbetrieb erreicht",
    28: "Raumsolltemperatur Absenkbetrieb erreicht",
    29: "Min. Rücklauftemperatur – Regelung vampair",
    30: "Außenabschalttemperatur Kühlen erreicht",
    31: "warte auf Kühlbetrieb der Wärmepumpe",
}

BUFFER_STATE = {
    0: "Status nicht vorhanden",
    1: "Bereitschaft",
    2: "Puffer wird beladen",
    3: "Frostschutzbetrieb",
    4: "Kaminkehrer",
    5: "Wärmeableitung",
    6: "Testlauf Pumpe ist aktiv ",
    7: "Trinkwasserspeicher wird beladen",
}

BUFFER_MODE = {0: "Immer Aus", 1: "Immer Ein", 2: "Zeitschaltung"}

BOILER_STATE = {
    0: "Boilerstatus nicht vorhanden",
    1: "Bereitschaft",
    2: "Laden",
    3: "Frostschutz",
    4: "Rauchfangkehrermodus",
    5: "Legionellenschutz",
    6: "Anforderung",
    7: "Energiequelle zu heiß",
    8: "Blockadeschutz",
    9: "einmalige Freigabe aktiv",
    10: "Fühler Kurzschluss",
    11: "Fühler Unterbrechung",
    12: "Ferienbetrieb",
    13: "Defrost",
}

BOILER_MODE = {
    0: "Immer Aus",
    1: "Immer Ein",
    2: "Montag – Sonntag",
    3: "Blockweise (Montag – Freitag, Samstag – Sonntag)",
    4: "Tagweise",
}

EVU_LOCK = {
    0: "Normalbetrieb",
    1: "EVU-Lock Aktiv",
}

DEFROST = {
    0: "Abtauung nicht aktiv",
    1: "Abtauung aktiv",
}

BOILER_CHARGE = {
    0: "Boilerladung nicht aktiv",
    1: "Boilerladung aktiv",
}

VAMPAIR_STATE = {
    0: "Bereitschaft",
    1: "Heizbetrieb",
    2: "Heizbetrieb, Trinkwasserspeicherladung",
    3: "Kühlbetrieb",
    4: "Manueller Betrieb",
    5: "EVU-Lock aktiv",
    6: "keine Zeitfreigabe, Wärmepumpe aus",
    7: "Außentemperatursperre, Wärmepumpe aus",
    8: "elektrische Zusatzheizung aktiv",
    9: "Fremdkessel aktiv, Wärmepumpe aus",
    10: "Kühlanforderung",
    11: "manuelle Leistungsvorgabe",
    12: "Wärmepumpe ausgeschaltet",
}

SMART_GRID_EINSCHALTUNG = 4
SMART_GRID_NORMALBETRIEB = 2



from .enums import RegisterTypes,DataTypes

class DataValue(object):
    """
    Abstraction of a certain relative address in the modbus register
    """
    type:DataTypes
    address:int
    count:int
    value:int
    multiplier:float|None
    _absolut_address:int
    register_type:RegisterTypes
    def __init__(self,address:int,count:int=1,default_value:int=0,multiplier:float=None,type:DataTypes=DataTypes.INT,register_type:RegisterTypes=RegisterTypes.Input) -> None:
        self.address = address
        self.count = count
        self.value = default_value
        self.multiplier = multiplier
        self.type = type
        self._absolut_address = None # This will be set by the parent component
        self.register_type = register_type
        
    def get_absolute_address(self)->int:
        """
        Returns the absolute address of the register
        """
        if not self._absolut_address:
            raise Exception("Absolute address not set!")
        return self._absolut_address + self.address  
    
    @property
    def has_scaler(self)->bool:
        return  self.multiplier is not None
    
    @property
    def scaled_value(self)->float:
        """
        Scaled value of this register
        """
        if self.has_scaler:
            return self.value * self.multiplier
        else:
            return self.value
       
    def reverse_scale(self,value:float)->float:
        """
        Applies the scaler in the reverse direction
        """
        if self.has_scaler:
            return value / self.multiplier
        else:
            return value
        
    def set_unscaled_value(self,value:float)->None:
        """
        Applies the scaler to the value and sets the value
        """
        self.value = self.reverse_scale(value)